from flask import Blueprint
from flask import render_template
import flask_login
import os as os

from ramp_database.model import Keyword
from ramp_database.model import Problem

from .redirect import redirect_to_user
from .._version import __version__

mod = Blueprint('general', __name__)


@mod.route('/')
def index():
    """Default landing page."""
    img_ext = ('.png', '.jpg', '.jpeg', '.gif', '.svg')
    current_dir = os.path.dirname(__file__)
    img_folder = os.path.join(current_dir, "..", "static", "img", "powered_by")
    context = {}
    if os.path.isdir(img_folder):
        images = [f for f in os.listdir(img_folder)
                  if f.endswith(img_ext)]
        context["images"] = images
    context["version"] = __version__
    return render_template('index.html', **context)


@mod.route("/description")
def ramp():
    """RAMP description request."""
    user = (flask_login.current_user
            if flask_login.current_user.is_authenticated else None)
    admin = user.access_level == 'admin' if user is not None else False
    return render_template('ramp_description.html', admin=admin)


@mod.route("/data_domains")
def data_domains():
    """Review of all possible keyword attached to the different RAMP
    problems."""
    current_keywords = Keyword.query.order_by(Keyword.name)
    current_problems = Problem.query.order_by(Problem.id)
    return render_template('data_domains.html',
                           keywords=current_keywords,
                           problems=current_problems)


@mod.route("/teaching")
def teaching():
    """Page related to RAMP offers for teaching classes."""
    return render_template('teaching.html')


@mod.route("/data_science_themes")
def data_science_themes():
    """Page reviewing problems organized by ML themes."""
    current_keywords = Keyword.query.order_by(Keyword.name)
    return render_template('data_science_themes.html',
                           keywords=current_keywords)


@mod.route("/keywords/<keyword_name>")
def keywords(keyword_name):
    """Page which give details about a keyword."""
    keyword = Keyword.query.filter_by(name=keyword_name).one_or_none()
    if keyword:
        return render_template('keyword.html', keyword=keyword)
    return redirect_to_user('Keyword {} does not exist.'
                            .format(keyword_name), is_error=True)



import numpy as np
import pandas as pd

from ramp_database.tools.submission import get_submission_by_id
from ramp_database.tools.submission import get_submissions


def make_step_df(pareto_df, is_lower_the_better):
    """Make a step function from pareto_df['x', 'y'].

    Parameters
    ----------
    pareto_df : pd.DataFrame
        Should contain 'x' and 'y' columns, sorted by 'x'.
    is_lower_the_better : boolean

    Returns
    ----------
    pareto_df : pd.DataFrame
    """
    n_pareto = len(pareto_df)
    pareto_df = pareto_df.set_index(1 + 2 * np.arange(n_pareto))
    for i in range(2, 2 * n_pareto, 2):
        pareto_df.loc[i] = pareto_df.loc[i - 1]
        pareto_df.loc[i, 'x'] = pareto_df.loc[i + 1, 'x']
    pareto_df.loc[2 * n_pareto] = pareto_df.loc[2 * n_pareto - 1]
    pareto_df.loc[2 * n_pareto, 'x'] = max(pareto_df['x'])
    pareto_df.loc[0] = pareto_df.loc[1]
    if is_lower_the_better:
        pareto_df.loc[0, 'y'] = max(pareto_df['y'])
    else:
        pareto_df.loc[0, 'y'] = min(pareto_df['y'])
    return pareto_df.sort_index()


def color_gradient(rgb, factor_array):
    """Rescale rgb by factor_array."""
    from skimage.color import gray2rgb, rgb2gray
    colors = np.array(
        [(255 - rgb[0], 255 - rgb[2], 255 - rgb[2]) for _ in factor_array])
    colors = rgb2gray(colors)
    colors = gray2rgb(255 - np.array([color * factor for color, factor
                                      in zip(colors, factor_array)]))[:, :, 0]
    return colors


def add_pareto(df, col, worst, is_lower_the_better):
    """Add a column col + ' pareto' with 1s where col is on Pareto front.

    It is assumed that the dataframe is oredered by the
    'x' variable that defines the Pareto front.

    Parameters
    ----------
    df : pd.DataFrame
        The dataframe to amend.
    col : string
        The numerical 'y' column.
    worst : number
        The worst possible value of col.
    is_lower_the_better : boolean

    Returns
    ----------
    df: pd.DataFrame
        The dataframe amended with the new column col + ' pareto'
    """
    df_ = df.copy()
    df_.loc[:, col + ' pareto'] = pd.Series(np.zeros(df.shape[0]),
                                            index=df_.index)
    best_score = worst
    if is_lower_the_better:
        for i, row in df.iterrows():
            score = row[col]
            if score < best_score:
                best_score = score
                df_.loc[i, col + ' pareto'] = 1
    else:
        for i, row in df.iterrows():
            score = row[col]
            if score > best_score:
                best_score = score
                df_.loc[i, col + ' pareto'] = 1
    return df_


def score_plot(session, event):
    from bokeh.plotting import figure
    from bokeh.models.sources import ColumnDataSource
    from bokeh.models.formatters import DatetimeTickFormatter

    submissions = get_submissions(session, event.name, None)
    submissions = [
        get_submission_by_id(session, sub_id)
        for sub_id, _, _ in submissions
        if get_submission_by_id(session, sub_id).is_public_leaderboard and
        get_submission_by_id(session, sub_id).is_valid]
    score_names = [score_type.name for score_type in event.score_types]
    scoress = np.array([
        [score.valid_score_cv_bag
         for score in submission.ordered_scores(score_names)]
        for submission in submissions
    ]).T

    score_plot_df = pd.DataFrame()
    score_plot_df['submitted at (UTC)'] = [
        submission.submission_timestamp for submission in submissions]
    score_plot_df['contributivity'] = [
        submission.contributivity for submission in submissions]
    score_plot_df['historical contributivity'] = [
        submission.historical_contributivity for submission in submissions]
    for score_name in score_names:  # to make sure the column is created
        score_plot_df[score_name] = 0
    for score_name, scores in zip(score_names, scoress):
        score_plot_df[score_name] = scores

    score_name = event.official_score_name
    score_plot_df = score_plot_df[
        score_plot_df['submitted at (UTC)'] > event.opening_timestamp]
    score_plot_df = score_plot_df.sort_values('submitted at (UTC)')
    score_plot_df = add_pareto(
        score_plot_df, score_name, event.official_score_type.worst,
        event.official_score_type.is_lower_the_better)

    is_open = (score_plot_df['submitted at (UTC)'] >
               event.public_opening_timestamp).values

    max_contributivity = max(
        0.0000001, max(score_plot_df['contributivity'].values))
    max_historical_contributivity = max(0.0000001, max(
        score_plot_df['historical contributivity'].values))

    fill_color_1 = (176, 23, 31)
    fill_color_2 = (16, 78, 139)
    fill_colors_1 = color_gradient(
        fill_color_1, score_plot_df['contributivity'].values /
        max_contributivity)
    fill_colors_2 = color_gradient(
        fill_color_2, score_plot_df['historical contributivity'].values /
        max_historical_contributivity)
    fill_colors = np.minimum(fill_colors_1, fill_colors_2).astype(int)
    fill_colors = ["#%02x%02x%02x" % (c[0], c[1], c[2]) for c in fill_colors]

    score_plot_df['x'] = score_plot_df['submitted at (UTC)']
    score_plot_df['y'] = score_plot_df[score_name]
    score_plot_df['line_color'] = 'royalblue'
    score_plot_df['circle_size'] = 8
    score_plot_df['line_color'] = 'royalblue'
    score_plot_df.loc[is_open, 'line_color'] = 'coral'
    score_plot_df['fill_color'] = fill_colors
    score_plot_df['fill_alpha'] = 0.5
    score_plot_df['line_width'] = 0
    score_plot_df['label'] = 'closed phase'
    score_plot_df.loc[is_open, 'label'] = 'open phase'

    source = ColumnDataSource(score_plot_df)
    pareto_df = score_plot_df[
        score_plot_df[score_name + ' pareto'] == 1].copy()
    pareto_df = pareto_df.append(pareto_df.iloc[-1])
    pareto_df.iloc[-1, pareto_df.columns.get_loc('x')] = (
        max(score_plot_df['x'])
    )
    pareto_df = make_step_df(
        pareto_df, event.official_score_type.is_lower_the_better)
    source_pareto = ColumnDataSource(pareto_df)

    tools = ['pan,wheel_zoom,box_zoom,reset,previewsave,tap']
    p = figure(plot_width=900, plot_height=600, tools=tools, title='Scores')

    p.circle(
        'x', 'y', size='circle_size', line_color='line_color',
        fill_color='fill_color', fill_alpha='fill_alpha', line_width=1,
        source=source, legend='label'
    )
    p.line(
        'x', 'y', line_width=3, line_color='goldenrod', source=source_pareto,
        legend='best score', alpha=0.9
    )

    p.xaxis.formatter = DatetimeTickFormatter(
        hours=['%d %B %Y'],
        days=['%d %B %Y'],
        months=['%d %B %Y'],
        years=['%d %B %Y'],
    )
    p.xaxis.major_label_orientation = np.pi / 4

    if event.official_score_type.is_lower_the_better:
        p.yaxis.axis_label = score_name + ' (the lower the better)'
        p.legend.location = 'top_right'
    else:
        p.yaxis.axis_label = score_name + ' (the greater the better)'
        p.legend.location = 'bottom_right'
    p.xaxis.axis_label = 'submission timestamp (UTC)'
    p.xaxis.axis_label_text_font_size = '14pt'
    p.yaxis.axis_label_text_font_size = '14pt'
    p.legend.label_text_font_size = '14pt'
    p.title.text_font_size = '16pt'
    p.xaxis.major_label_text_font_size = '10pt'
    p.yaxis.major_label_text_font_size = '10pt'
    return p



import shutil

import pytest

from ramp_utils import generate_flask_config
from ramp_utils import read_config
from ramp_utils.testing import database_config_template
from ramp_utils.testing import ramp_config_template

from ramp_database.model import Model
from ramp_database.testing import create_toy_db
from ramp_database.utils import setup_db
from ramp_database.utils import session_scope

from ramp_frontend import create_app


@pytest.fixture(scope='module')
def client_session(database_connection):
    database_config = read_config(database_config_template())
    ramp_config = ramp_config_template()
    try:
        deployment_dir = create_toy_db(database_config, ramp_config)
        flask_config = generate_flask_config(database_config)
        app = create_app(flask_config)
        app.config['TESTING'] = True
        app.config['WTF_CSRF_ENABLED'] = False
        with session_scope(database_config['sqlalchemy']) as session:
            yield app.test_client(), session
    finally:
        shutil.rmtree(deployment_dir, ignore_errors=True)
        try:
            # In case of failure we should close the global flask engine
            from ramp_frontend import db as db_flask
            db_flask.session.close()
        except RuntimeError:
            pass
        db, _ = setup_db(database_config['sqlalchemy'])
        Model.metadata.drop_all(db)


def test_index(client_session):
    client, _ = client_session
    rv = client.get('/')
    assert rv.status_code == 200
    assert (b'RAMP: collaborative data science challenges' in
            rv.data)


def test_ramp(client_session):
    client, _ = client_session
    rv = client.get('/description')
    assert rv.status_code == 200
    assert (b'The RAMP software packages were developed by the' in
            rv.data)


def test_domain(client_session):
    client, session = client_session
    rv = client.get('/data_domains')
    assert rv.status_code == 200
    assert b'Scientific data domains' in rv.data
    assert b'boston_housing' in rv.data
    assert b'Boston housing price regression' in rv.data


def test_teaching(client_session):
    client, _ = client_session
    rv = client.get('/teaching')
    assert rv.status_code == 200
    assert b'RAMP challenges begin with an interesting supervised prediction' \
        in rv.data


def test_data_science_themes(client_session):
    client, _ = client_session
    rv = client.get('/data_science_themes')
    assert rv.status_code == 200
    assert b'boston_housing_theme' in rv.data
    assert b'iris_theme' in rv.data


def test_keywords(client_session):
    client, _ = client_session
    rv = client.get('/keywords/boston_housing')
    assert rv.status_code == 200
    assert b'Related problems' in rv.data
    assert b'boston_housing' in rv.data
    assert b'Boston housing price regression' in rv.data



import re
import shutil

import pytest

from werkzeug.datastructures import ImmutableMultiDict

from ramp_utils import generate_flask_config
from ramp_utils import read_config
from ramp_utils.testing import database_config_template
from ramp_utils.testing import ramp_config_template

from ramp_database.model import Model
from ramp_database.testing import create_toy_db
from ramp_database.utils import setup_db
from ramp_database.utils import session_scope

from ramp_database.tools.event import get_event
from ramp_database.tools.user import add_user
from ramp_database.tools.user import get_user_by_name
from ramp_database.tools.team import ask_sign_up_team
from ramp_database.tools.team import get_event_team_by_name

from ramp_frontend import create_app
from ramp_frontend.testing import login_scope


@pytest.fixture(scope='module')
def client_session(database_connection):
    database_config = read_config(database_config_template())
    ramp_config = ramp_config_template()
    try:
        deployment_dir = create_toy_db(database_config, ramp_config)
        flask_config = generate_flask_config(database_config)
        app = create_app(flask_config)
        app.config['TESTING'] = True
        app.config['WTF_CSRF_ENABLED'] = False
        with session_scope(database_config['sqlalchemy']) as session:
            yield app.test_client(), session
    finally:
        shutil.rmtree(deployment_dir, ignore_errors=True)
        try:
            # In case of failure we should close the global flask engine
            from ramp_frontend import db as db_flask
            db_flask.session.close()
        except RuntimeError:
            pass
        db, _ = setup_db(database_config['sqlalchemy'])
        Model.metadata.drop_all(db)


@pytest.mark.parametrize(
    "page",
    ["/approve_users",
     "/sign_up/test_user",
     "/events/iris_test/sign_up/test_user",
     "/events/iris_test/update",
     "/user_interactions",
     "/events/iris_test/dashboard_submissions"]
)
def test_check_login_required(client_session, page):
    client, _ = client_session

    rv = client.get(page)
    assert rv.status_code == 302
    assert 'http://localhost/login' in rv.location
    rv = client.get(page, follow_redirects=True)
    assert rv.status_code == 200


@pytest.mark.parametrize(
    "page, request_function",
    [("/approve_users", ["get", "post"]),
     ("/sign_up/test_user", ["get"]),
     ("/events/iris_test/sign_up/test_user", ["get"]),
     ("/events/iris_test/update", ["get", "post"]),
     ("/user_interactions", ["get"]),
     ("/events/iris_test/dashboard_submissions", ["get"])]
)
def test_check_admin_required(client_session, page, request_function):
    client, _ = client_session

    with login_scope(client, 'test_user', 'test') as client:
        for rf in request_function:
            rv = getattr(client, rf)(page)
            with client.session_transaction() as cs:
                flash_message = dict(cs['_flashes'])
            assert (flash_message['message'] ==
                    'Sorry User, you do not have admin rights')
            assert rv.status_code == 302
            assert rv.location == 'http://localhost/problems'
            rv = getattr(client, rf)(page, follow_redirects=True)
            assert rv.status_code == 200


def test_approve_users_remove(client_session):
    client, session = client_session

    # create 2 new users
    add_user(session, 'xx', 'xx', 'xx', 'xx', 'xx', access_level='user')
    add_user(session, 'yy', 'yy', 'yy', 'yy', 'yy', access_level='asked')
    # ask for sign up for an event for the first user
    _, _, event_team = ask_sign_up_team(session, 'iris_test', 'xx')

    with login_scope(client, 'test_iris_admin', 'test') as client:

        # GET check that we get all new user to be approved
        rv = client.get('/approve_users')
        assert rv.status_code == 200
        # line for user approval
        assert b'yy yy - yy' in rv.data
        # line for the event approval
        assert b'iris_test - xx'

        # POST check that we are able to approve a user and event
        data = ImmutableMultiDict([
            ('submit_button', 'Remove!'),
            ('approve_users', 'yy'),
            ('approve_event_teams', str(event_team.id))
        ])
        rv = client.post('/approve_users', data=data)
        assert rv.status_code == 302
        assert rv.location == 'http://localhost/problems'

        # ensure that the previous change have been committed within our
        # session
        session.commit()
        user = get_user_by_name(session, 'yy')
        assert user is None
        event_team = get_event_team_by_name(session, 'iris_test', 'xx')
        assert event_team is None
        with client.session_transaction() as cs:
            flash_message = dict(cs['_flashes'])
        assert re.match(r"Removed users:\nyy\nRemoved event_team:\n"
                        r"Event\(iris_test\)/Team\(.*xx.*\)\n",
                        flash_message['Removed users'])


def test_approve_users_approve(client_session):
    client, session = client_session

    # create 2 new users
    add_user(session, 'cc', 'cc', 'cc', 'cc', 'cc', access_level='user')
    add_user(session, 'dd', 'dd', 'dd', 'dd', 'dd', access_level='asked')
    # ask for sign up for an event for the first user
    _, _, event_team = ask_sign_up_team(session, 'iris_test', 'cc')

    with login_scope(client, 'test_iris_admin', 'test') as client:

        # GET check that we get all new user to be approved
        rv = client.get('/approve_users')
        assert rv.status_code == 200
        # line for user approval
        assert b'dd dd - dd' in rv.data
        # line for the event approval
        assert b'iris_test - cc'

        # POST check that we are able to approve a user and event
        data = ImmutableMultiDict([
            ('submit_button', 'Approve!'),
            ('approve_users', 'dd'),
            ('approve_event_teams', str(event_team.id))]
        )
        rv = client.post('/approve_users', data=data)
        assert rv.status_code == 302
        assert rv.location == 'http://localhost/problems'

        # ensure that the previous change have been committed within our
        # session
        session.commit()
        user = get_user_by_name(session, 'dd')
        assert user.access_level == 'user'
        event_team = get_event_team_by_name(session, 'iris_test', 'cc')
        assert event_team.approved
        with client.session_transaction() as cs:
            flash_message = dict(cs['_flashes'])
        assert re.match(r"Approved users:\ndd\nApproved event_team:\n"
                        r"Event\(iris_test\)/Team\(.*cc.*\)\n",
                        flash_message['Approved users'])


def test_approve_single_user(client_session):
    client, session = client_session

    add_user(session, 'gg', 'gg', 'gg', 'gg', 'gg', access_level='asked')
    with login_scope(client, 'test_iris_admin', 'test') as client:
        rv = client.get('/sign_up/gg')
        assert rv.status_code == 302
        assert rv.location == 'http://localhost/problems'
        with client.session_transaction() as cs:
            flash_message = dict(cs['_flashes'])
        assert re.match("User(.*gg.*) is signed up",
                        flash_message['Successful sign-up'])

        # ensure that the previous change have been committed within our
        # session
        session.commit()
        user = get_user_by_name(session, 'gg')
        assert user.access_level == 'user'

        rv = client.get("/sign_up/unknown_user")
        session.commit()
        assert rv.status_code == 302
        assert rv.location == "http://localhost/problems"
        with client.session_transaction() as cs:
            flash_message = dict(cs['_flashes'])
        assert flash_message['message'] == 'No user unknown_user'


def test_approve_sign_up_for_event(client_session):
    client, session = client_session

    with login_scope(client, 'test_iris_admin', 'test') as client:

        # check the redirection if the user or the event does not exist
        rv = client.get("/events/xxx/sign_up/test_user")
        session.commit()
        assert rv.status_code == 302
        assert rv.location == "http://localhost/problems"
        with client.session_transaction() as cs:
            flash_message = dict(cs['_flashes'])
        assert flash_message['message'] == 'No event xxx or no user test_user'

        rv = client.get("/events/iris_test/sign_up/xxxx")
        session.commit()
        assert rv.status_code == 302
        assert rv.location == "http://localhost/problems"
        with client.session_transaction() as cs:
            flash_message = dict(cs['_flashes'])
        assert flash_message['message'] == 'No event iris_test or no user xxxx'

        add_user(session, 'zz', 'zz', 'zz', 'zz', 'zz', access_level='user')
        _, _, event_team = ask_sign_up_team(session, 'iris_test', 'zz')
        assert not event_team.approved
        rv = client.get('/events/iris_test/sign_up/zz')
        assert rv.status_code == 302
        assert rv.location == "http://localhost/problems"
        session.commit()
        event_team = get_event_team_by_name(session, 'iris_test', 'zz')
        assert event_team.approved
        with client.session_transaction() as cs:
            flash_message = dict(cs['_flashes'])
        assert "is signed up for Event" in flash_message['Successful sign-up']


def test_update_event(client_session):
    client, session = client_session

    with login_scope(client, 'test_iris_admin', 'test') as client:

        # case tha the event does not exist
        rv = client.get('/events/boston_housing/update')
        assert rv.status_code == 302
        assert rv.location == 'http://localhost/problems'
        with client.session_transaction() as cs:
            flash_message = dict(cs['_flashes'])
        assert 'no event named "boston_housing"' in flash_message['message']

        # GET: pre-fill the forms
        rv = client.get('/events/iris_test/update')
        assert rv.status_code == 200
        assert b'Minimum duration between submissions' in rv.data

        # POST: update the event data
        event_info = {
            'suffix': 'test',
            'title': 'Iris new title',
            'is_send_trained_mail': True,
            'is_public': True,
            'is_controled_signup': True,
            'is_competitive': False,
            'min_duration_between_submissions_hour': 0,
            'min_duration_between_submissions_minute': 0,
            'min_duration_between_submissions_second': 0,
            'opening_timestamp': "2000-01-01 00:00:00",
            'closing_timestamp': "2100-01-01 00:00:00",
            'public_opening_timestamp': "2000-01-01 00:00:00",
        }
        rv = client.post('/events/iris_test/update', data=event_info)
        assert rv.status_code == 302
        assert rv.location == "http://localhost/problems"
        event = get_event(session, 'iris_test')
        assert event.min_duration_between_submissions == 0


def test_user_interactions(client_session):
    client, _ = client_session

    with login_scope(client, 'test_iris_admin', 'test') as client:
        rv = client.get('/user_interactions')
        assert rv.status_code == 200
        assert b'landing' in rv.data


# TODO: To be tested when we implemented properly the leaderboard
# def test_dashboard_submissions(client_session):
#     client, session = client_session

#     with login_scope(client, 'test_iris_admin', 'test') as client:
#         rv = client.get('/events/iris_test/dashboard_submissions')
#         print(rv.data.decode('utf-8'))



import os
import shutil
import signal
import subprocess
import time

import pytest

from ramp_utils import read_config
from ramp_utils.testing import database_config_template
from ramp_utils.testing import ramp_config_template

from ramp_database.utils import setup_db
from ramp_database.model import Model
from ramp_database.testing import create_toy_db


@pytest.fixture(scope="module")
def make_toy_db(database_connection):
    database_config = read_config(database_config_template())
    ramp_config = ramp_config_template()
    try:
        deployment_dir = create_toy_db(database_config, ramp_config)
        yield
    finally:
        shutil.rmtree(deployment_dir, ignore_errors=True)
        db, _ = setup_db(database_config['sqlalchemy'])
        Model.metadata.drop_all(db)


def test_test_launch(make_toy_db):
    # pass environment to subprocess
    cmd = ['python', '-m']
    cmd += ["ramp_frontend.cli", "test-launch",
            "--config", database_config_template()]
    proc = subprocess.Popen(
        cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
        env=os.environ.copy())
    # wait for 5 seconds before to terminate the server
    time.sleep(5)
    proc.send_signal(signal.SIGINT)
    stdout, _ = proc.communicate()
    assert b'Serving Flask app "ramp-frontend"' in stdout



"""
The :mod:`ramp_frontend.utils` provides utilities to ease sending email.
"""

import logging

from flask_mail import Message

from ramp_frontend import mail

logger = logging.getLogger('RAMP-FRONTEND')


def body_formatter_user(user):
    """Create the body of an email using the user information.

    Parameters
    ----------
    user : :class:`ramp_database.model.User`
        The user profile.

    Returns
    -------
    body : str
        The email body.
    """
    body = """
    user = {}
    name = {} {}
    email = {}
    linkedin = {}
    twitter = {}
    facebook = {}
    github = {}
    notes = {}
    bio = {}

    """.format(user.name, user.firstname,
               user.lastname, user.email, user.linkedin_url,
               user.twitter_url, user.facebook_url, user.github_url,
               user.hidden_notes, user.bio)

    return body


def send_mail(to, subject, body):
    """Send email using Flask Mail.

    Parameters
    ----------
    to : str
        The email address of the recipient.
    subject : str
        The subject of the email.
    body : str
        The body of the email.
    """
    try:
        msg = Message(subject)
        msg.body = body
        msg.add_recipient(to)
        mail.send(msg)
    except Exception as e:
        logger.error('Mailing error: {}'.format(e))



"""
The :mod:`ramp_frontend.forms` module defines the different forms used on the
website.
"""

from flask_wtf import FlaskForm

from wtforms import BooleanField
from wtforms import DateField
from wtforms import DateTimeField
from wtforms import FileField
from wtforms import IntegerField
from wtforms import PasswordField
from wtforms import SelectMultipleField
from wtforms import StringField
from wtforms import validators
from wtforms import ValidationError
from wtforms.widgets import CheckboxInput
from wtforms.widgets import ListWidget


def _space_check(form, field):
    if ' ' in field.data:
        raise ValidationError('Field cannot contain space.')


def _ascii_check(form, field):
    try:
        field.data.encode('ascii')
    except Exception:
        raise ValidationError('Field cannot contain non-ascii characters.')


class LoginForm(FlaskForm):
    """Login-in form.

    Attributes
    ----------
    user_name : str
        The user name.
    password : str
        The user password.
    """
    user_name = StringField('user_name', [validators.DataRequired()])
    password = PasswordField('password', [validators.DataRequired()])


class UserUpdateProfileForm(FlaskForm):
    """User profile form.

    Attributes
    ----------
    user_name : str
        The user name.
    firstname : str
        The user's first name.
    lastname : str
        The user's last name.
    email : str
        The user's email address.
    linkedin_url : str, default == ''
        The user's LinkedIn URL.
    twitter_url : str, defaut == ''
        The user's Twitter URL.
    facebook_url : str, default == ''
        The user's Facebook URL.
    google_url : str, default == ''
        The user's Google URL.
    github_url : str, default == ''
        The user's GitHub URL.
    website_url : str, default == ''
        The user's website URL.
    bio : str, default == ''
        The user's bio.
    is_want_news : bool, default is True
        Whether the user want some info from us.
    """
    user_name = StringField('user_name', [
        validators.DataRequired(), validators.Length(min=1, max=20),
        _space_check])
    firstname = StringField('firstname', [validators.DataRequired()])
    lastname = StringField('lastname', [validators.DataRequired()])
    email = StringField('email', [validators.DataRequired()])
    linkedin_url = StringField('linkedin_url')
    twitter_url = StringField('twitter_url')
    facebook_url = StringField('facebook_url')
    google_url = StringField('google_url')
    github_url = StringField('github_url')
    website_url = StringField('website_url')
    bio = StringField('bio')
    is_want_news = BooleanField()


class UserCreateProfileForm(UserUpdateProfileForm):
    """User profile form.

    Attributes
    ----------
    user_name : str
        The user name.
    password : str
        The user password.
    firstname : str
        The user's first name.
    lastname : str
        The user's last name.
    email : str
        The user's email address.
    linkedin_url : str, default == ''
        The user's LinkedIn URL.
    twitter_url : str, defaut == ''
        The user's Twitter URL.
    facebook_url : str, default == ''
        The user's Facebook URL.
    google_url : str, default == ''
        The user's Google URL.
    github_url : str, default == ''
        The user's GitHub URL.
    website_url : str, default == ''
        The user's website URL.
    bio : str, default == ''
        The user's bio.
    is_want_news : bool, default is True
        Whether the user want some info from us.
    """
    password = PasswordField('password', [validators.DataRequired()])


class CodeForm(FlaskForm):
    """Code form.

    This is the form used to contain the code when submitting to RAMP.

    Attributes
    ----------
    named_codes : list of tuple (submission_file_name, submission_code)
        The place holder containing the name of the submission file and the
        code associated.
    """
    names_codes = []


class SubmitForm(FlaskForm):
    """Submission name form.

    This is the form where the name of the submission given by the user will
    be stored.

    Attributes
    ----------
    submission_name : str
        The name of the submission.
    """
    submission_name = StringField('submission_name',
                                  [validators.DataRequired(), _space_check])


class UploadForm(FlaskForm):
    """Submission uploading form.

    This is the form used to upload a file to be loaded during a RAMP
    submission.

    Attributes
    ----------
    file : file
        File to be uploaded and loaded into the sandbox code form.
    """
    file = FileField('file')


class EventUpdateProfileForm(FlaskForm):
    """Form to update the parameters of an event.

    Attributes
    ----------
    title : str
        The event title.
    is_send_trained_mails : bool
        Whether or not to send an email when submissions are trained.
    is_public : bool
        Whether or not the event is public.
    is_controled_signup : bool
        Whether or not the event has controlled sign-up.
    is_competitive : bool
        Whether or not the event has a competitive phase.
    min_duration_between_submission_hour : int
        The number of hour to wait between two submissions.
    min_duration_between_submission_minute : int
        The number of minute to wait between two submissions.
    min_duration_between_submission_second : int
        The number of second to wait between two submissions.
    opening_timestamp : datetime
        The date and time when the event is opening.
    closing_timestamp : datetime
        The date and time when the event is closing.
    public_opening_timestamp : datetime
        The date and time when the public phase of the event is opening.
    """
    title = StringField(
        'event_title', [validators.DataRequired(), validators.Length(max=80)]
    )
    is_send_trained_mails = BooleanField()
    is_send_submitted_mails = BooleanField()
    is_public = BooleanField()
    is_controled_signup = BooleanField()
    is_competitive = BooleanField()
    min_duration_between_submissions_hour = IntegerField(
        'min_h', [validators.NumberRange(min=0)]
    )
    min_duration_between_submissions_minute = IntegerField(
        'min_m', [validators.NumberRange(min=0, max=59)]
    )
    min_duration_between_submissions_second = IntegerField(
        'min_s', [validators.NumberRange(min=0, max=59)]
    )
    opening_timestamp = DateTimeField(
        'opening_timestamp', [], format='%Y-%m-%d %H:%M:%S'
    )
    closing_timestamp = DateTimeField(
        'closing_timestamp', [], format='%Y-%m-%d %H:%M:%S'
    )
    public_opening_timestamp = DateTimeField(
        'public_opening_timestamp', [], format='%Y-%m-%d %H:%M:%S'
    )


class MultiCheckboxField(SelectMultipleField):
    """A form containing multiple checkboxes."""
    widget = ListWidget(prefix_label=False)
    option_widget = CheckboxInput()


class ImportForm(FlaskForm):
    """The form allowing to select which model to view."""
    selected_f_names = MultiCheckboxField('selected_f_names')


class CreditForm(FlaskForm):
    """Credit form.

    The credit form is used to acknowledge other submission when making a
    RAMP submission after tracking the user activity.

    Attributes
    ----------
    note : str
        Some notes regarding the credit.
    self_credit : str
        The credit given to the current submission.
    name_credits : list
        The name for the credits.
    """
    note = StringField('submission_name')
    self_credit = StringField('self credit')
    name_credits = []


class AskForEventForm(FlaskForm):
    """Form to ask for a new event.

    Attributes
    ----------
    suffix : str
        The suffix used for the event.
    title : str
        The event title.
    n_students : int
        The number of students that will take part in the event.
    min_duration_between_submission_hour : int
        The number of hour to wait between two submissions.
    min_duration_between_submission_minute : int
        The number of minute to wait between two submissions.
    min_duration_between_submission_second : int
        The number of second to wait between two submissions.
    opening_timestamp : datetime
        The date and time when the event is opening.
    closing_timestamp : datetime
        The date and time when the event is closing.
    """
    suffix = StringField(
        'event_suffix',
        [validators.DataRequired(), validators.Length(max=20), _ascii_check,
         _space_check]
    )
    title = StringField(
        'event_title',
        [validators.DataRequired(), validators.Length(max=80)]
    )
    n_students = IntegerField(
        'n_students',
        [validators.DataRequired(), validators.NumberRange(min=0)]
    )
    min_duration_between_submissions_hour = IntegerField(
        'min_h', [validators.NumberRange(min=0)]
    )
    min_duration_between_submissions_minute = IntegerField(
        'min_m', [validators.NumberRange(min=0, max=59)]
    )
    min_duration_between_submissions_second = IntegerField(
        'min_s', [validators.NumberRange(min=0, max=59)]
    )
    opening_date = DateField(
        'opening_date', [validators.DataRequired()], format='%Y-%m-%d'
    )
    closing_date = DateField(
        'closing_date', [validators.DataRequired()], format='%Y-%m-%d'
    )


class EmailForm(FlaskForm):
    email = StringField(
        'Email', validators=[validators.DataRequired(), validators.Email()]
    )


class PasswordForm(FlaskForm):
    password = PasswordField(
        'Password', validators=[validators.DataRequired()]
    )



"""Blueprint for all authentication functions for the RAMP frontend."""
import logging

import flask_login

from flask import abort
from flask import Blueprint
from flask import current_app as app
from flask import flash
from flask import redirect
from flask import request
from flask import render_template
from flask import session
from flask import url_for

from itsdangerous import URLSafeTimedSerializer

from ramp_database.utils import check_password
from ramp_database.utils import hash_password

from ramp_database.tools.user import add_user
from ramp_database.tools.user import add_user_interaction
from ramp_database.tools.user import get_user_by_name_or_email
from ramp_database.tools.user import set_user_by_instance

from ramp_database.model import User

from ramp_database.exceptions import NameClashError

from ramp_frontend import db
from ramp_frontend import login_manager

from ..forms import EmailForm
from ..forms import LoginForm
from ..forms import PasswordForm
from ..forms import UserCreateProfileForm
from ..forms import UserUpdateProfileForm

from ..utils import body_formatter_user
from ..utils import send_mail

logger = logging.getLogger('RAMP-FRONTEND')
mod = Blueprint('auth', __name__)
ts = URLSafeTimedSerializer(app.config["SECRET_KEY"])


@login_manager.user_loader
def load_user(id):
    """Load a user in the login manager.

    This function is used by Flask-Login to manage the current-user connection.

    Parameters
    ----------
    id : int
        The user ID.
    """
    return User.query.get(id)


@mod.route("/login", methods=['GET', 'POST'])
def login():
    """Login request."""
    if app.config['TRACK_USER_INTERACTION']:
        add_user_interaction(db.session, interaction='landing')

    if flask_login.current_user.is_authenticated:
        logger.info('User already logged-in')
        session['logged_in'] = True
        return redirect(url_for('ramp.problems'))

    form = LoginForm()
    if form.validate_on_submit():
        user = get_user_by_name_or_email(db.session,
                                         name=form.user_name.data)
        if user is None:
            msg = 'User "{}" does not exist'.format(form.user_name.data)
            flash(msg)
            logger.info(msg)
            return redirect(url_for('auth.login'))
        if not check_password(form.password.data,
                              user.hashed_password):
            msg = 'Wrong password'
            flash(msg)
            logger.info(msg)
            return redirect(url_for('auth.login'))
        flask_login.login_user(user, remember=True)
        session['logged_in'] = True
        user.is_authenticated = True
        db.session.commit()
        logger.info('User "{}" is logged in'
                    .format(flask_login.current_user.name))
        if app.config['TRACK_USER_INTERACTION']:
            add_user_interaction(
                db.session, interaction='login', user=flask_login.current_user
            )
        next_ = request.args.get('next')
        if next_ is None:
            next_ = url_for('ramp.problems')
        return redirect(next_)

    return render_template('login.html', form=form)


@mod.route("/logout")
@flask_login.login_required
def logout():
    """Logout request."""
    user = flask_login.current_user
    if app.config['TRACK_USER_INTERACTION']:
        add_user_interaction(db.session, interaction='logout', user=user)
    session['logged_in'] = False
    user.is_authenticated = False
    db.session.commit()
    logger.info('{} is logged out'.format(user))
    flask_login.logout_user()

    return redirect(url_for('auth.login'))


@mod.route("/sign_up", methods=['GET', 'POST'])
def sign_up():
    """Sign-up request."""
    if flask_login.current_user.is_authenticated:
        session['logged_in'] = True
        return redirect(url_for('ramp.problems'))

    form = UserCreateProfileForm()
    if form.validate_on_submit():
        try:
            user = add_user(
                session=db.session,
                name=form.user_name.data,
                password=form.password.data,
                lastname=form.lastname.data,
                firstname=form.firstname.data,
                email=form.email.data,
                linkedin_url=form.linkedin_url.data,
                twitter_url=form.twitter_url.data,
                facebook_url=form.facebook_url.data,
                google_url=form.google_url.data,
                github_url=form.github_url.data,
                website_url=form.website_url.data,
                bio=form.bio.data,
                is_want_news=form.is_want_news.data,
                access_level='not_confirmed'
            )
        except NameClashError as e:
            flash(str(e))
            logger.info(str(e))
            return redirect(url_for('auth.sign_up'))
        # send an email to the participant such that he can confirm his email
        token = ts.dumps(user.email)
        recover_url = url_for(
            'auth.user_confirm_email', token=token, _external=True
        )
        subject = "Confirm your email for signing-up to RAMP"
        body = ('Hi {}, \n\n Click on the following link to confirm your email'
                ' address and finalize your sign-up to RAMP.\n\n Note that '
                'your account still needs to be approved by a RAMP '
                'administrator.\n\n'
                .format(user.firstname))
        body += recover_url
        body += '\n\nSee you on the RAMP website!'
        send_mail(user.email, subject, body)
        logger.info(
            '{} has signed-up to RAMP'.format(user.name)
        )
        flash(
            "We sent a confirmation email. Go read your email and click on "
            "the confirmation link"
        )
        return redirect(url_for('auth.login'))
    return render_template('sign_up.html', form=form)


@mod.route("/update_profile", methods=['GET', 'POST'])
@flask_login.login_required
def update_profile():
    """User profile update."""
    form = UserUpdateProfileForm()
    form.user_name.data = flask_login.current_user.name
    if form.validate_on_submit():
        set_user_by_instance(
            db.session,
            user=flask_login.current_user,
            lastname=form.lastname.data,
            firstname=form.firstname.data,
            email=form.email.data,
            linkedin_url=form.linkedin_url.data,
            twitter_url=form.twitter_url.data,
            facebook_url=form.facebook_url.data,
            google_url=form.google_url.data,
            github_url=form.github_url.data,
            website_url=form.website_url.data,
            is_want_news=form.is_want_news.data
        )
        # send_register_request_mail(user)
        return redirect(url_for('ramp.problems'))
    form.lastname.data = flask_login.current_user.lastname
    form.firstname.data = flask_login.current_user.firstname
    form.email.data = flask_login.current_user.email
    form.linkedin_url.data = flask_login.current_user.linkedin_url
    form.twitter_url.data = flask_login.current_user.twitter_url
    form.facebook_url.data = flask_login.current_user.facebook_url
    form.google_url.data = flask_login.current_user.google_url
    form.github_url.data = flask_login.current_user.github_url
    form.website_url.data = flask_login.current_user.website_url
    form.bio.data = flask_login.current_user.bio
    form.is_want_news.data = flask_login.current_user.is_want_news
    return render_template('update_profile.html', form=form)


@mod.route('/reset_password', methods=["GET", "POST"])
def reset_password():
    """Reset password of a RAMP user."""
    form = EmailForm()
    error = ''
    if form.validate_on_submit():
        user = User.query.filter_by(email=form.email.data).one_or_none()
        if user and user.access_level != 'asked':
            token = ts.dumps(user.email)
            recover_url = url_for(
                'auth.reset_with_token', token=token, _external=True
            )

            subject = "Password reset requested - RAMP website"
            body = ('Hi {}, \n\nclick on the link to reset your password:\n'
                    .format(user.firstname))
            body += recover_url
            body += '\n\nSee you on the RAMP website!'
            send_mail(user.email, subject, body)
            logger.info(
                'Password reset requested for user {}'.format(user.name)
            )
            logger.info(recover_url)
            flash('An email to reset your password has been sent')
            return redirect(url_for('auth.login'))
        elif user is None:
            error = ('The email address is not linked to any user. You can '
                     'sign-up instead.')
        else:
            error = ('Your account has not been yet approved. You cannot '
                     'change the password already.')
    return render_template('reset_password.html', form=form, error=error)


@mod.route('/reset/<token>', methods=["GET", "POST"])
def reset_with_token(token):
    """Reset password by passing a token (email).

    Parameters
    ----------
    token : str
        The token associated with an email address.
    """
    try:
        email = ts.loads(token, max_age=86400)
    except Exception as e:
        logger.error(str(e))
        abort(404)

    form = PasswordForm()
    if form.validate_on_submit():
        user = User.query.filter_by(email=email).one_or_none()
        if user is None:
            logger.error('The error was deleted before resetting his/her '
                         'password')
            abort(404)
        (User.query.filter_by(email=email)
                   .update({
                       "hashed_password":
                       hash_password(form.password.data).decode()}))
        db.session.commit()
        return redirect(url_for('auth.login'))

    return render_template('reset_with_token.html', form=form, token=token)


@mod.route('/confirm_email/<token>', methods=["GET", "POST"])
def user_confirm_email(token):
    """Confirm a user account using his email address and a token to approve.

    Parameters
    ----------
    token : str
        The token associated with an email address.
    """
    try:
        email = ts.loads(token, max_age=86400)
    except Exception as e:
        logger.error(str(e))
        abort(404)

    user = User.query.filter_by(email=email).one_or_none()
    if user is None:
        flash(
            'You did not sign-up yet to RAMP. Please sign-up first.',
            category='error'
        )
        return redirect(url_for('auth.sign_up'))
    elif user.access_level in ('user', 'admin'):
        flash(
            "Your account is already approved. You don't need to confirm your "
            "email address", category='error'
        )
        return redirect(url_for('auth.login'))
    elif user.access_level == 'asked':
        flash(
            "Your email address already has been confirmed. You need to wait "
            "for an approval from a RAMP administrator", category='error'
        )
        return redirect(url_for('general.index'))
    User.query.filter_by(email=email).update({'access_level': 'asked'})
    db.session.commit()
    admin_users = User.query.filter_by(access_level='admin')
    for admin in admin_users:
        subject = 'Approve registration of {}'.format(
            user.name
        )
        body = body_formatter_user(user)
        url_approve = ('http://{}/sign_up/{}'
                       .format(app.config['DOMAIN_NAME'], user.name))
        body += 'Click on the link to approve the registration '
        body += 'of this user: {}'.format(url_approve)
        send_mail(admin.email, subject, body)
    flash(
        "An email has been sent to the RAMP administrator(s) who will "
        "approve your account"
    )
    return redirect(url_for('auth.login'))



"""The :mod:`ramp_frontend.testing` module contains all functions used to
easily test the frontend."""

import errno
import socket
from contextlib import contextmanager

import pytest


def login(client, username, password):
    """Simulate a log-in from a user.

    After the log-in request, the client will be redirect to the expected page.

    Parameters
    ----------
    client : :class:`flask.testing.FlaskClient`
        The testing client used for unit testing.
    username : str
        The user's name.
    password : str
        The user's password.

    Returns
    -------
    response : :class:`flask.wrappers.Response`
        The response of the client.
    """
    return client.post('/login', data=dict(
        user_name=username,
        password=password
    ), follow_redirects=True)


def logout(client):
    """Simulate a log-out.

    After the log-out request, the client will be redirected to the expected
    page.

    Parameters
    ----------
    client : :class:`flask.testing.FlaskClient`
        The testing client used for unit testing.

    Returns
    -------
    response : :class:`flask.wrappers.Response`
        The response of the client.
    """
    return client.get('/logout', follow_redirects=True)


@contextmanager
def login_scope(client, username, password):
    """Context manager to log-in during the ``with`` scope.

    Parameters
    ----------
    client : :class:`flask.testing.FlaskClient`
        The testing client used for unit testing.
    username : str
        The user's name.
    password : str
        The user's password.

    Returns
    -------
    client : :class:`flask.testing.FlaskClient`
        A client which is logged-in for the duration of the ``with`` scope.
    """
    login(client, username, password)
    yield client
    logout(client)


def _bind_smtp_port():
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    port_in_use = False

    try:
        s.bind(("127.0.0.1", 8025))
    except socket.error as e:
        if e.errno == errno.EADDRINUSE:
            port_in_use = True

    s.close()
    return port_in_use


_fail_no_smtp_server = pytest.mark.skipif(
    not _bind_smtp_port(), reason="No smtp server in use."
)



import logging

from flask import flash
from flask import redirect
from flask import url_for

logger = logging.getLogger('RAMP-FRONTEND')


def redirect_to_user(message_str, is_error=True, category='message'):
    """Redirect the page to the problem landing page.

    Parameters
    ----------
    message_str : str
        The message to inform the user what is going on.
    is_error : bool, default is True
        Whether this is due to an error.
    category : str or None
        The category of the error. Refer to :func:`flask.flash`.
    """
    flash(message_str, category=category)
    print(message_str)
    if is_error:
        logger.error(message_str)
    else:
        logger.info(message_str)
    return redirect(url_for('ramp.problems'))


def redirect_to_sandbox(event, message_str, is_error=True, category=None):
    """Redirect the page to the sandbox landing page.

    Parameters
    ----------
    message_str : str
        The message to inform the user what is going on.
    is_error : bool, default is True
        Whether this is due to an error.
    category : str or None
        The category of the error. Refer to :func:`flask.flash`.
    """
    flash(message_str, category=category)
    if is_error:
        logger.error(message_str)
    else:
        logger.info(message_str)
    return redirect('/events/{}/sandbox'.format(event.name))


def redirect_to_credit(submission_hash, message_str, is_error=True,
                       category=None):
    """Redirect the page to the credit landing page.

    Parameters
    ----------
    submission_hash : str
        The hash of the current submission.
    message_str : str
        The message to inform the user what is going on.
    is_error : bool, default is True
        Whether this is due to an error.
    category : str or None
        The category of the error. Refer to :func:`flask.flash`.
    """
    flash(message_str, category=category)
    if is_error:
        logger.error(message_str)
    else:
        logger.info(message_str)
    return redirect('/credit/{}'.format(submission_hash))



import datetime
import difflib
import logging
import io
import os
import shutil
import tempfile
import time
import zipfile

from bokeh.embed import components

import flask_login

from flask import Blueprint
from flask import current_app as app
from flask import redirect
from flask import render_template
from flask import request
from flask import send_from_directory
from flask import send_file

from wtforms import StringField
from wtforms.widgets import TextArea

from werkzeug.utils import secure_filename

from ramp_database.model import Event
from ramp_database.model import Problem
from ramp_database.model import Submission
from ramp_database.model import SubmissionFile
from ramp_database.model import SubmissionSimilarity
from ramp_database.model import User
from ramp_database.model import WorkflowElement

from ramp_database.exceptions import DuplicateSubmissionError
from ramp_database.exceptions import MissingExtensionError
from ramp_database.exceptions import TooEarlySubmissionError

from ramp_database.tools.event import get_event
from ramp_database.tools.event import get_problem
from ramp_database.tools.frontend import is_admin
from ramp_database.tools.frontend import is_accessible_code
from ramp_database.tools.frontend import is_accessible_event
from ramp_database.tools.frontend import is_user_signed_up
from ramp_database.tools.frontend import is_user_sign_up_requested
from ramp_database.tools.leaderboard import update_leaderboards
from ramp_database.tools.submission import add_submission
from ramp_database.tools.submission import add_submission_similarity
from ramp_database.tools.submission import get_source_submissions
from ramp_database.tools.submission import get_submission_by_name
from ramp_database.tools.user import add_user_interaction
from ramp_database.tools.team import ask_sign_up_team
from ramp_database.tools.team import get_event_team_by_name
from ramp_database.tools.team import sign_up_team

from ramp_frontend import db

from ..forms import AskForEventForm
from ..forms import CodeForm
from ..forms import CreditForm
from ..forms import ImportForm
from ..forms import SubmitForm
from ..forms import UploadForm

from ..utils import body_formatter_user
from ..utils import send_mail

from .redirect import redirect_to_credit
from .redirect import redirect_to_sandbox
from .redirect import redirect_to_user

from .visualization import score_plot

mod = Blueprint('ramp', __name__)
logger = logging.getLogger('RAMP-FRONTEND')


@mod.route("/problems")
def problems():
    """Landing page showing all the RAMP problems."""
    user = (flask_login.current_user
            if flask_login.current_user.is_authenticated else None)
    admin = user.access_level == 'admin' if user is not None else False
    if app.config['TRACK_USER_INTERACTION']:
        add_user_interaction(
            db.session, interaction='looking at problems', user=user
        )
    problems = get_problem(db.session, None)

    for problem in problems:
        for event in problem.events:
            # check the state of the event
            now = datetime.datetime.now()
            start = event.opening_timestamp
            start_collab = event.public_opening_timestamp
            end = event.closing_timestamp
            if now < start or now >= end:
                event.state = 'close'
            elif now >= start and now < start_collab:
                event.state = 'competitive'
            elif now >= start and now >= start_collab and now < end:
                event.state = 'collab'
            if user:
                signed = get_event_team_by_name(
                                    db.session, event.name,
                                    flask_login.current_user.name)
                if not signed:
                    event.state_user = 'not_signed'
                elif signed.approved:
                    event.state_user = 'signed'
                elif signed:
                    event.state_user = 'waiting'
            else:
                event.state_user = 'not_signed'

    # problems = Problem.query.order_by(Problem.id.desc())
    return render_template('problems.html',
                           problems=problems,
                           admin=admin)


@mod.route("/problems/<problem_name>")
def problem(problem_name):
    """Landing page for a single RAMP problem.

    Parameters
    ----------
    problem_name : str
        The name of a problem.
    """
    current_problem = get_problem(db.session, problem_name)
    user = (flask_login.current_user
            if flask_login.current_user.is_authenticated else None)
    admin = user.access_level == 'admin' if user is not None else False
    if current_problem:
        if app.config['TRACK_USER_INTERACTION']:
            if flask_login.current_user.is_authenticated:
                add_user_interaction(
                    db.session,
                    interaction='looking at problem',
                    user=flask_login.current_user,
                    problem=current_problem
                )
            else:
                add_user_interaction(
                    db.session, interaction='looking at problem',
                    problem=current_problem
                )
        description_f_name = os.path.join(
            current_problem.path_ramp_kit,
            '{}_starting_kit.html'.format(current_problem.name)
        )
        # check which event ramp-kit archive is the latest
        archive_dir = os.path.join(
            current_problem.path_ramp_kit, "events_archived"
        )
        latest_event_zip = max(
            [f for f in os.scandir(archive_dir) if f.name.endswith(".zip")],
            key=lambda x: x.stat().st_mtime
        )
        latest_event = os.path.splitext(latest_event_zip.name)[0]

        return render_template(
            'problem.html', problem=current_problem, admin=admin,
            notebook_filename=description_f_name, latest_event=latest_event
        )
    else:
        return redirect_to_user('Problem {} does not exist'
                                .format(problem_name), is_error=True)


@mod.route("/download_starting_kit/<event_name>")
def download_starting_kit(event_name):
    event = db.session.query(Event).filter_by(name=event_name).one()
    return send_from_directory(
        os.path.join(event.problem.path_ramp_kit, "events_archived"),
        event_name + ".zip"
    )


@mod.route("/notebook/<problem_name>")
def notebook(problem_name):
    current_problem = get_problem(db.session, problem_name)
    return send_from_directory(
        current_problem.path_ramp_kit,
        '{}_starting_kit.html'.format(current_problem.name)
    )


@mod.route("/rules/<event_name>")
def rules(event_name):
    event = get_event(db.session, event_name)
    return render_template('rules.html', event=event)


@mod.route("/events/<event_name>")
@flask_login.login_required
def user_event(event_name):
    """Landing page for a given event.

    Parameters
    ----------
    event_name : str
        The event name.
    """
    if flask_login.current_user.access_level == 'asked':
        msg = 'Your account has not been approved yet by the administrator'
        logger.error(msg)
        return redirect_to_user(msg)
    if not is_accessible_event(db.session, event_name,
                               flask_login.current_user.name):
        return redirect_to_user('{}: no event named "{}"'
                                .format(flask_login.current_user.firstname,
                                        event_name))
    event = get_event(db.session, event_name)
    if event:
        if app.config['TRACK_USER_INTERACTION']:
            add_user_interaction(db.session, interaction='looking at event',
                                 event=event, user=flask_login.current_user)
        admin = is_admin(db.session, event_name, flask_login.current_user.name)
        approved = is_user_signed_up(
            db.session, event_name, flask_login.current_user.name
        )
        asked = is_user_sign_up_requested(
            db.session, event_name, flask_login.current_user.name
        )
        return render_template('event.html',
                               event=event,
                               admin=admin,
                               approved=approved,
                               asked=asked)
    return redirect_to_user('Event {} does not exist.'
                            .format(event_name), is_error=True)


@mod.route("/events/<event_name>/sign_up")
@flask_login.login_required
def sign_up_for_event(event_name):
    """Landing page to sign-up to a specific RAMP event.

    Parameters
    ----------
    event_name : str
        The name of the event.
    """
    event = get_event(db.session, event_name)
    if not is_accessible_event(db.session, event_name,
                               flask_login.current_user.name):
        return redirect_to_user('{}: no event named "{}"'
                                .format(flask_login.current_user.firstname,
                                        event_name))
    if app.config['TRACK_USER_INTERACTION']:
        add_user_interaction(db.session, interaction='signing up at event',
                             user=flask_login.current_user, event=event)

    ask_sign_up_team(db.session, event.name, flask_login.current_user.name)
    if event.is_controled_signup:
        admin_users = User.query.filter_by(access_level='admin')
        for admin in admin_users:
            subject = ('Request to sign-up {} to RAMP event {}'
                       .format(event.name, flask_login.current_user.name))
            body = body_formatter_user(flask_login.current_user)
            url_approve = ('http://{}/events/{}/sign_up/{}'
                           .format(
                               app.config['DOMAIN_NAME'], event.name,
                               flask_login.current_user.name
                           ))
            body += ('Click on this link to approve the sign-up request: {}'
                     .format(url_approve))
            send_mail(admin.email, subject, body)
        return redirect_to_user("Sign-up request is sent to event admins.",
                                is_error=False, category='Request sent')
    sign_up_team(db.session, event.name, flask_login.current_user.name)
    return redirect_to_sandbox(
        event,
        '{} is signed up for {}.'
        .format(flask_login.current_user.firstname, event),
        is_error=False,
        category='Successful sign-up'
    )


@mod.route("/events/<event_name>/sandbox", methods=['GET', 'POST'])
@flask_login.login_required
def sandbox(event_name):
    """Landing page for the user's sandbox.

    Parameters
    ----------
    event_name : str
        The event name.
    """
    event = get_event(db.session, event_name)
    if not is_accessible_event(db.session, event_name,
                               flask_login.current_user.name):
        return redirect_to_user(
            '{}: no event named "{}"'
            .format(flask_login.current_user.firstname, event_name)
        )
    if not is_accessible_code(db.session, event_name,
                              flask_login.current_user.name):
        error_str = ('No access to sandbox for event {}. If you have '
                     'already signed up, please wait for approval.'
                     .format(event.name))
        return redirect_to_user(error_str)
    # setup the webpage when loading
    # we use the code store in the sandbox to show to the user
    sandbox_submission = get_submission_by_name(
        db.session, event_name, flask_login.current_user.name,
        event.ramp_sandbox_name
    )
    event_team = get_event_team_by_name(
        db.session, event_name, flask_login.current_user.name
    )
    # initialize the form for the code
    # The amount of python magic we have to do for rendering a variable
    # number of textareas, named and populated at run time, is mind
    # boggling.

    # First we need to make sure CodeForm is empty
    # for name_code in CodeForm.names_codes:
    #     name, _ = name_code
    #     delattr(CodeForm, name)
    CodeForm.names_codes = []

    # Then we create named fields in the CodeForm class for each editable
    # submission file. They have to be populated when the code_form object
    # is created, so we also create a code_form_kwargs dictionary and
    # populate it with the codes.
    code_form_kwargs = {}
    for submission_file in sandbox_submission.files:
        if submission_file.is_editable:
            f_field = submission_file.name
            setattr(CodeForm,
                    f_field, StringField('Text', widget=TextArea()))
            code_form_kwargs[f_field] = submission_file.get_code()
    code_form_kwargs['prefix'] = 'code'
    code_form = CodeForm(**code_form_kwargs)
    # Then, to be able to iterate over the files in the sandbox.html
    # template, we also fill a separate table of pairs (file name, code).
    # The text areas in the template will then have to be created manually.
    for submission_file in sandbox_submission.files:
        if submission_file.is_editable:
            code_form.names_codes.append(
                (submission_file.name, submission_file.get_code()))

    # initialize the submission field and the the uploading form
    submit_form = SubmitForm(
        submission_name=event_team.last_submission_name, prefix='submit'
    )
    upload_form = UploadForm(prefix='upload')

    #  check if the event is before, during or after open state
    now = datetime.datetime.now()
    start = event.opening_timestamp
    end = event.closing_timestamp

    event_status = {"msg": "",
                    "state": "not_yet"}
    start_str = start.strftime("%d of %B %Y at %H:%M")
    end_str = end.strftime("%d of %B %Y, %H:%M")
    if now < start:
        event_status["msg"] = "Event submissions will open on the " + start_str
        event_status["state"] = "close"
    elif now < end:
        event_status["msg"] = "Event submissions are open until " + end_str
        event_status["state"] = "open"
    else:  # now >= end
        event_status["msg"] = "This event closed on the " + end_str
        event_status["state"] = "close"

    admin = is_admin(db.session, event_name, flask_login.current_user.name)
    if request.method == 'GET':
        return render_template(
            'sandbox.html',
            submission_names=sandbox_submission.f_names,
            code_form=code_form,
            submit_form=submit_form, upload_form=upload_form,
            event=event,
            admin=admin,
            event_status=event_status
        )

    if request.method == 'POST':
        if ('code-csrf_token' in request.form and
                code_form.validate_on_submit()):
            try:
                for submission_file in sandbox_submission.files:
                    if submission_file.is_editable:
                        old_code = submission_file.get_code()
                        submission_file.set_code(
                            request.form[submission_file.name])
                        new_code = submission_file.get_code()
                        diff = '\n'.join(difflib.unified_diff(
                            old_code.splitlines(), new_code.splitlines()))
                        similarity = difflib.SequenceMatcher(
                            a=old_code, b=new_code).ratio()
                        if app.config['TRACK_USER_INTERACTION']:
                            add_user_interaction(
                                db.session,
                                interaction='save',
                                user=flask_login.current_user,
                                event=event,
                                submission_file=submission_file,
                                diff=diff, similarity=similarity
                            )
            except Exception as e:
                return redirect_to_sandbox(event, 'Error: {}'.format(e))

            # if we required to only save the file, redirect now
            if "saving" in request.form:
                return redirect_to_sandbox(
                    event,
                    'Your submission has been saved. You can safely comeback '
                    'to your sandbox later.',
                    is_error=False, category='File saved'
                )

        elif request.files:
            upload_f_name = secure_filename(
                request.files['file'].filename)
            upload_name = upload_f_name.split('.')[0]
            # TODO: create a get_function
            upload_workflow_element = WorkflowElement.query.filter_by(
                name=upload_name, workflow=event.workflow).one_or_none()
            if upload_workflow_element is None:
                return redirect_to_sandbox(event,
                                           '{} is not in the file list.'
                                           .format(upload_f_name))

            # TODO: create a get_function
            submission_file = SubmissionFile.query.filter_by(
                submission=sandbox_submission,
                workflow_element=upload_workflow_element).one()
            if submission_file.is_editable:
                old_code = submission_file.get_code()

            tmp_f_name = os.path.join(tempfile.gettempdir(), upload_f_name)
            request.files['file'].save(tmp_f_name)
            file_length = os.stat(tmp_f_name).st_size
            if (upload_workflow_element.max_size is not None and
                    file_length > upload_workflow_element.max_size):
                return redirect_to_sandbox(
                    event,
                    'File is too big: {} exceeds max size {}'
                    .format(file_length, upload_workflow_element.max_size)
                )
            if submission_file.is_editable:
                try:
                    with open(tmp_f_name) as f:
                        code = f.read()
                        submission_file.set_code(code)
                except Exception as e:
                    return redirect_to_sandbox(event, 'Error: {}'.format(e))
            else:
                # non-editable files are not verified for now
                dst = os.path.join(sandbox_submission.path, upload_f_name)
                shutil.copy2(tmp_f_name, dst)
            logger.info('{} uploaded {} in {}'
                        .format(flask_login.current_user.name, upload_f_name,
                                event))

            if submission_file.is_editable:
                new_code = submission_file.get_code()
                diff = '\n'.join(difflib.unified_diff(
                    old_code.splitlines(), new_code.splitlines()))
                similarity = difflib.SequenceMatcher(
                    a=old_code, b=new_code).ratio()
                if app.config['TRACK_USER_INTERACTION']:
                    add_user_interaction(
                        db.session,
                        interaction='upload',
                        user=flask_login.current_user,
                        event=event,
                        submission_file=submission_file,
                        diff=diff,
                        similarity=similarity
                    )
            else:
                if app.config['TRACK_USER_INTERACTION']:
                    add_user_interaction(
                        db.session,
                        interaction='upload',
                        user=flask_login.current_user,
                        event=event,
                        submission_file=submission_file
                    )

            return redirect(request.referrer)
            # TODO: handle different extensions for the same workflow element
            # ie: now we let upload eg external_data.bla, and only fail at
            # submission, without giving a message

        if 'submission' in request.form:
            if not submit_form.validate_on_submit():
                return redirect_to_sandbox(
                    event,
                    'Submission name should not contain any spaces'
                )
            new_submission_name = request.form['submit-submission_name']
            if not 4 < len(new_submission_name) < 20:
                return redirect_to_sandbox(
                    event,
                    'Submission name should have length between 4 and '
                    '20 characters.'
                )
            try:
                new_submission_name.encode('ascii')
            except Exception as e:
                return redirect_to_sandbox(event, 'Error: {}'.format(e))
            try:
                new_submission = add_submission(db.session, event_name,
                                                event_team.team.name,
                                                new_submission_name,
                                                sandbox_submission.path)
            except DuplicateSubmissionError:
                return redirect_to_sandbox(
                    event,
                    'Submission {} already exists. Please change the name.'
                    .format(new_submission_name)
                )

            except MissingExtensionError:
                return redirect_to_sandbox(
                    event, 'Missing extension'
                )
            except TooEarlySubmissionError as e:
                return redirect_to_sandbox(event, str(e))

            logger.info('{} submitted {} for {}.'
                        .format(flask_login.current_user.name,
                                new_submission.name, event_team))
            if event.is_send_submitted_mails:
                admin_users = User.query.filter_by(access_level='admin')
                for admin in admin_users:
                    subject = 'Submission {} sent for training'.format(
                        new_submission.name
                    )
                    body = """A new submission have been submitted:
                    event: {}
                    user: {}
                    submission: {}
                    submission path: {}
                    """.format(event_team.event.name,
                               flask_login.current_user.name,
                               new_submission.name, new_submission.path)
                    send_mail(admin.email, subject, body)
            if app.config['TRACK_USER_INTERACTION']:
                add_user_interaction(
                    db.session,
                    interaction='submit',
                    user=flask_login.current_user,
                    event=event,
                    submission=new_submission
                )

            return redirect_to_sandbox(
                event,
                '{} submitted {} for {}'
                .format(flask_login.current_user.firstname,
                        new_submission.name, event_team),
                is_error=False, category='Submission'
            )

    admin = is_admin(db.session, event_name, flask_login.current_user.name)
    return render_template(
        'sandbox.html',
        submission_names=sandbox_submission.f_names,
        code_form=code_form,
        submit_form=submit_form, upload_form=upload_form,
        event=event,
        admin=admin,
        event_status=event_status
    )


@mod.route("/problems/<problem_name>/ask_for_event", methods=['GET', 'POST'])
@flask_login.login_required
def ask_for_event(problem_name):
    problem = Problem.query.filter_by(name=problem_name).one_or_none()
    if problem is None:
        return redirect_to_user(
            '{}: no problem named "{}"'
            .format(flask_login.current_user.firstname, problem_name)
        )
    logger.info('{} is asking for event on {}'
                .format(flask_login.current_user.name, problem.name))
    # We assume here that event name has the syntax <problem_name>_<suffix>
    form = AskForEventForm(
        min_duration_between_submissions_hour=8,
        min_duration_between_submissions_minute=0,
        min_duration_between_submissions_second=0,
    )
    if form.validate_on_submit():
        admin_users = User.query.filter_by(access_level='admin')
        for admin in admin_users:
            subject = 'Request to add a new event'
            body = """User {} asked to add a new event:
            event name: {}
            event title: {}
            number of students: {}
            waiting time between resubmission: {}:{}:{}
            opening data: {}
            closing data: {}
            """.format(
                flask_login.current_user.name,
                problem.name + '_' + form.suffix.data,
                form.title.data,
                form.n_students.data,
                form.min_duration_between_submissions_hour.data,
                form.min_duration_between_submissions_minute.data,
                form.min_duration_between_submissions_second.data,
                form.opening_date.data,
                form.closing_date.data
            )
            send_mail(admin.email, subject, body)
        return redirect_to_user(
            'Thank you. Your request has been sent to RAMP administrators.',
            category='Event request', is_error=False
        )

    return render_template('ask_for_event.html', form=form, problem=problem)


@mod.route("/credit/<submission_hash>", methods=['GET', 'POST'])
@flask_login.login_required
def credit(submission_hash):
    """The landing page to credit other submission when a user submit is own.

    Parameters
    ----------
    submission_hash : str
        The submission hash of the current submission.
    """
    submission = (Submission.query.filter_by(hash_=submission_hash)
                                  .one_or_none())
    access_code = is_accessible_code(
        db.session, submission.event_team.event.name,
        flask_login.current_user.name, submission.id
    )
    if submission is None or not access_code:
        error_str = 'Missing submission: {}'.format(submission_hash)
        return redirect_to_user(error_str)
    event_team = submission.event_team
    event = event_team.event
    source_submissions = get_source_submissions(db.session, submission.id)

    def get_s_field(source_submission):
        return '{}/{}/{}'.format(
            source_submission.event_team.event.name,
            source_submission.event_team.team.name,
            source_submission.name)

    # Make sure that CreditForm is empty
    CreditForm.name_credits = []
    credit_form_kwargs = {}
    for source_submission in source_submissions:
        s_field = get_s_field(source_submission)
        setattr(CreditForm, s_field, StringField('Text'))
    credit_form = CreditForm(**credit_form_kwargs)
    sum_credit = 0
    # new = True
    for source_submission in source_submissions:
        s_field = get_s_field(source_submission)
        submission_similaritys = \
            (SubmissionSimilarity.query
                                 .filter_by(
                                     type='target_credit',
                                     user=flask_login.current_user,
                                     source_submission=source_submission,
                                     target_submission=submission)
                                 .all())
        if not submission_similaritys:
            submission_credit = 0
        else:
            # new = False
            # find the last credit (in case crediter changes her mind)
            submission_similaritys.sort(
                key=lambda x: x.timestamp, reverse=True)
            submission_credit = int(
                round(100 * submission_similaritys[0].similarity)
            )
            sum_credit += submission_credit
        credit_form.name_credits.append(
            (s_field, str(submission_credit), source_submission.link)
        )
    # This doesnt work, not sure why
    # if not new:
    #    credit_form.self_credit.data = str(100 - sum_credit)
    if credit_form.validate_on_submit():
        try:
            sum_credit = int(credit_form.self_credit.data)
            logger.info(sum_credit)
            for source_submission in source_submissions:
                s_field = get_s_field(source_submission)
                sum_credit += int(getattr(credit_form, s_field).data)
            if sum_credit != 100:
                return redirect_to_credit(
                    submission_hash,
                    'Error: The total credit should add up to 100'
                )
        except Exception as e:
            return redirect_to_credit(submission_hash, 'Error: {}'.format(e))
        for source_submission in source_submissions:
            s_field = get_s_field(source_submission)
            similarity = int(getattr(credit_form, s_field).data) / 100.
            submission_similarity = \
                (SubmissionSimilarity.query
                                     .filter_by(
                                         type='target_credit',
                                         user=flask_login.current_user,
                                         source_submission=source_submission,
                                         target_submission=submission)
                                     .all())
            # if submission_similarity is not empty, we need to
            # add zero to cancel previous credits explicitly
            if similarity > 0 or submission_similarity:
                add_submission_similarity(
                    db.session,
                    credit_type='target_credit',
                    user=flask_login.current_user,
                    source_submission=source_submission,
                    target_submission=submission,
                    similarity=similarity,
                    timestamp=datetime.datetime.utcnow()
                )

        if app.config['TRACK_USER_INTERACTION']:
            add_user_interaction(
                db.session,
                interaction='giving credit',
                user=flask_login.current_user,
                event=event,
                submission=submission
            )

        return redirect('/events/{}/sandbox'.format(event.name))

    admin = is_admin(db.session, event.name, flask_login.current_user.name)
    return render_template(
        'credit.html', submission=submission,
        source_submissions=source_submissions, credit_form=credit_form,
        event=event, admin=admin)


@mod.route("/event_plots/<event_name>")
@flask_login.login_required
def event_plots(event_name):
    """Landing page of the plot illustrating the score evolution over time for
    a specific RAMP event.

    Parameters
    ----------
    event_name : str
        The name of the event.
    """
    event = get_event(db.session, event_name)
    if not is_accessible_event(db.session, event_name,
                               flask_login.current_user.name):
        return redirect_to_user('{}: no event named "{}"'
                                .format(flask_login.current_user.firstname,
                                        event_name))
    if event:
        p = score_plot(db.session, event)
        script, div = components(p)
        return render_template('event_plots.html',
                               script=script,
                               div=div,
                               event=event)
    return redirect_to_user('Event {} does not exist.'
                            .format(event_name),
                            is_error=True)


@mod.route("/<submission_hash>/<f_name>", methods=['GET', 'POST'])
@flask_login.login_required
def view_model(submission_hash, f_name):
    """Rendering submission codes using templates/submission.html.

    The code of f_name is displayed in the left panel, the list of submissions
    files is in the right panel. Clicking on a file will show that file (using
    the same template). Clicking on the name on the top will download the file
    itself (managed in the template). Clicking on "Archive" will zip all the
    submission files and download them (managed here).

    Parameters
    ----------
    submission_hash : str
        The hash_ of the submission.
    f_name : tr
        The name of the submission file.
    """
    submission = (Submission.query.filter_by(hash_=submission_hash)
                                  .one_or_none())
    if (submission is None or
            not is_accessible_code(db.session, submission.event.name,
                                   flask_login.current_user.name,
                                   submission.id)):
        error_str = 'Missing submission: {}'.format(submission_hash)
        return redirect_to_user(error_str)
    event = submission.event_team.event
    team = submission.event_team.team
    workflow_element_name = f_name.split('.')[0]
    workflow_element = \
        (WorkflowElement.query.filter_by(name=workflow_element_name,
                                         workflow=event.workflow)
                              .one_or_none())
    if workflow_element is None:
        error_str = ('{} is not a valid workflow element by {} '
                     .format(workflow_element_name,
                             flask_login.current_user.name))
        error_str += 'in {}/{}/{}/{}'.format(event, team, submission, f_name)
        return redirect_to_user(error_str)
    submission_file = \
        (SubmissionFile.query.filter_by(submission=submission,
                                        workflow_element=workflow_element)
                             .one_or_none())
    if submission_file is None:
        error_str = ('No submission file by {} in {}/{}/{}/{}'
                     .format(flask_login.current_user.name,
                             event, team, submission, f_name))
        return redirect_to_user(error_str)

    # superfluous, perhaps when we'll have different extensions?
    f_name = submission_file.f_name

    submission_abspath = os.path.abspath(submission.path)
    if not os.path.exists(submission_abspath):
        error_str = ('{} does not exist by {} in {}/{}/{}/{}'
                     .format(submission_abspath, flask_login.current_user.name,
                             event, team, submission, f_name))
        return redirect_to_user(error_str)

    if app.config['TRACK_USER_INTERACTION']:
        add_user_interaction(
            db.session,
            interaction='looking at submission',
            user=flask_login.current_user,
            event=event,
            submission=submission,
            submission_file=submission_file
        )

    logger.info('{} is looking at {}/{}/{}/{}'
                .format(flask_login.current_user.name, event, team, submission,
                        f_name))

    # Downloading file if it is not editable (e.g., external_data.csv)
    if not workflow_element.is_editable:
        # archive_filename = f_name  + '.zip'
        # with changedir(submission_abspath):
        #    with ZipFile(archive_filename, 'w') as archive:
        #        archive.write(f_name)
        if app.config['TRACK_USER_INTERACTION']:
            add_user_interaction(
                db.session,
                interaction='download',
                user=flask_login.current_user,
                event=event,
                submission=submission,
                submission_file=submission_file
            )

        return send_from_directory(
            submission_abspath, f_name, as_attachment=True,
            attachment_filename='{}_{}'.format(submission.hash_[:6], f_name),
            mimetype='application/octet-stream'
        )

    # Importing selected files into sandbox
    choices = [(f, f) for f in submission.f_names]
    import_form = ImportForm()
    import_form.selected_f_names.choices = choices
    if import_form.validate_on_submit():
        sandbox_submission = get_submission_by_name(
            db.session, event.name, flask_login.current_user.name,
            event.ramp_sandbox_name
        )
        for filename in import_form.selected_f_names.data:
            logger.info(
                '{} is importing {}/{}/{}/{}'
                .format(flask_login.current_user.name, event, team,
                        submission, filename)
            )

            workflow_element = WorkflowElement.query.filter_by(
                name=filename.split('.')[0], workflow=event.workflow).one()

            # TODO: deal with different extensions of the same file
            src = os.path.join(submission.path, filename)
            dst = os.path.join(sandbox_submission.path, filename)
            shutil.copy2(src, dst)  # copying also metadata
            logger.info('Copying {} to {}'.format(src, dst))

            submission_file = SubmissionFile.query.filter_by(
                submission=submission,
                workflow_element=workflow_element).one()
            if app.config['TRACK_USER_INTERACTION']:
                add_user_interaction(
                    db.session,
                    interaction='copy',
                    user=flask_login.current_user,
                    event=event,
                    submission=submission,
                    submission_file=submission_file
                )

        return redirect('/events/{}/sandbox'.format(event.name))

    with open(os.path.join(submission.path, f_name)) as f:
        code = f.read()
    admin = is_admin(db.session, event.name, flask_login.current_user.name)
    return render_template(
        'submission.html',
        event=event,
        code=code,
        submission=submission,
        f_name=f_name,
        import_form=import_form,
        admin=admin)


@mod.route("/<submission_hash>/error.txt")
@flask_login.login_required
def view_submission_error(submission_hash):
    """Rendering submission codes using templates/submission.html.

    The code of f_name is displayed in the left panel, the list of submissions
    files is in the right panel. Clicking on a file will show that file (using
    the same template). Clicking on the name on the top will download the file
    itself (managed in the template). Clicking on "Archive" will zip all the
    submission files and download them (managed here).

    Parameters
    ----------
    submission_hash : str
        The hash of the submission.
    """
    submission = (Submission.query.filter_by(hash_=submission_hash)
                                  .one_or_none())
    if submission is None:
        error_str = ('Missing submission {}: {}'
                     .format(flask_login.current_user.name, submission_hash))
        return redirect_to_user(error_str)
    event = submission.event_team.event
    team = submission.event_team.team
    # TODO: check if event == submission.event_team.event

    if app.config['TRACK_USER_INTERACTION']:
        add_user_interaction(
            db.session,
            interaction='looking at error',
            user=flask_login.current_user,
            event=event,
            submission=submission
        )

    return render_template(
        'submission_error.html', submission=submission, team=team, event=event
    )


@mod.route("/toggle_competition/<submission_hash>")
@flask_login.login_required
def toggle_competition(submission_hash):
    """Pulling out or putting a submission back into competition.

    Parameters
    ----------
    submission_hash : str
        The submission hash of the current submission.
    """
    submission = (Submission.query.filter_by(hash_=submission_hash)
                                  .one_or_none())
    if submission is None:
        error_str = 'Missing submission: {}'.format(submission_hash)
        return redirect_to_user(error_str)

    access_code = is_accessible_code(
        db.session, submission.event_team.event.name,
        flask_login.current_user.name, submission.id
    )
    if not access_code:
        error_str = 'Missing submission: {}'.format(submission_hash)
        return redirect_to_user(error_str)

    submission.is_in_competition = not submission.is_in_competition
    db.session.commit()
    update_leaderboards(db.session, submission.event_team.event.name)
    return redirect(
        '/{}/{}'.format(submission_hash, submission.files[0].f_name)
    )


@mod.route("/download/<submission_hash>")
@flask_login.login_required
def download_submission(submission_hash):
    """Download a submission from the server.

    Parameters
    ----------
    submission_hash : str
        The submission hash of the current submission.
    """
    submission = (Submission.query.filter_by(hash_=submission_hash)
                                  .one_or_none())
    if submission is None:
        error_str = 'Missing submission: {}'.format(submission_hash)
        return redirect_to_user(error_str)

    access_code = is_accessible_code(
        db.session, submission.event_team.event.name,
        flask_login.current_user.name, submission.id
    )
    if not access_code:
        error_str = 'Unauthorized access: {}'.format(submission_hash)
        return redirect_to_user(error_str)

    file_in_memory = io.BytesIO()
    with zipfile.ZipFile(file_in_memory, 'w') as zf:
        for ff in submission.files:
            data = zipfile.ZipInfo(ff.f_name)
            data.date_time = time.localtime(time.time())[:6]
            data.compress_type = zipfile.ZIP_DEFLATED
            zf.writestr(data, ff.get_code())
    file_in_memory.seek(0)
    return send_file(
        file_in_memory,
        attachment_filename=f"submission_{submission.id}.zip",
        as_attachment=True,
    )



import shutil

import pytest

from ramp_utils import generate_flask_config
from ramp_utils import read_config
from ramp_utils.testing import database_config_template
from ramp_utils.testing import ramp_config_template

from ramp_database.model import Model
from ramp_database.testing import create_toy_db
from ramp_database.utils import setup_db
from ramp_database.utils import session_scope

from ramp_database.tools.user import get_user_by_name

from ramp_frontend import create_app
from ramp_frontend import mail
from ramp_frontend.utils import body_formatter_user
from ramp_frontend.utils import send_mail
from ramp_frontend.testing import _fail_no_smtp_server


@pytest.fixture(scope='module')
def client_session(database_connection):
    database_config = read_config(database_config_template())
    ramp_config = ramp_config_template()
    try:
        deployment_dir = create_toy_db(database_config, ramp_config)
        flask_config = generate_flask_config(database_config)
        app = create_app(flask_config)
        app.config['TESTING'] = True
        app.config['WTF_CSRF_ENABLED'] = False
        with session_scope(database_config['sqlalchemy']) as session:
            yield app.test_client(), session
    finally:
        shutil.rmtree(deployment_dir, ignore_errors=True)
        try:
            # In case of failure we should close the global flask engine
            from ramp_frontend import db as db_flask
            db_flask.session.close()
        except RuntimeError:
            pass
        db, _ = setup_db(database_config['sqlalchemy'])
        Model.metadata.drop_all(db)


@_fail_no_smtp_server
def test_send_mail(client_session):
    client, _ = client_session
    with client.application.app_context():
        with mail.record_messages() as outbox:
            send_mail('xx@gmail.com', 'subject', 'body')
            assert len(outbox) == 1
            assert outbox[0].subject == 'subject'
            assert outbox[0].body == 'body'
            assert outbox[0].recipients == ['xx@gmail.com']


def test_body_formatter_user(client_session):
    _, session = client_session
    user = get_user_by_name(session, 'test_user')
    for word in ['test_user', 'User', 'Test', 'linkedin', 'twitter',
                 'facebook', 'github', 'notes', 'bio']:
        assert word in body_formatter_user(user)



#! /usr/bin/env python
import codecs
import os

from setuptools import find_packages, setup

# get __version__ from _version.py
ver_file = os.path.join('ramp_frontend', '_version.py')
with open(ver_file) as f:
    exec(f.read())

DISTNAME = 'ramp-frontend'
DESCRIPTION = "Website for RAMP"
with codecs.open('README.rst', encoding='utf-8-sig') as f:
    LONG_DESCRIPTION = f.read()
MAINTAINER = 'A. Boucaud, B. Kegl, G. Lemaitre, J. Van den Bossche'
MAINTAINER_EMAIL = 'boucaud.alexandre@gmail.com, guillaume.lemaitre@inria.fr'
URL = 'https://github.com/paris-saclay-cds/ramp-board'
LICENSE = 'BSD (3-clause)'
DOWNLOAD_URL = 'https://github.com/paris-saclay-cds/ramp-board'
VERSION = __version__  # noqa
CLASSIFIERS = ['Intended Audience :: Science/Research',
               'Intended Audience :: Developers',
               'License :: OSI Approved',
               'Programming Language :: Python',
               'Topic :: Software Development',
               'Topic :: Scientific/Engineering',
               'Operating System :: Microsoft :: Windows',
               'Operating System :: POSIX',
               'Operating System :: Unix',
               'Operating System :: MacOS',
               'Programming Language :: Python :: 3.6',
               'Programming Language :: Python :: 3.7']
INSTALL_REQUIRES = ['bokeh', 'click', 'Flask', 'Flask-Login', 'Flask-Mail',
                    'Flask-SQLAlchemy', 'Flask-WTF', 'itsdangerous', 'numpy',
                    'pandas']
EXTRAS_REQUIRE = {
    'tests': ['pytest', 'pytest-cov'],
    'docs': ['sphinx', 'sphinx_rtd_theme', 'numpydoc']
}
PACKAGE_DATA = {
    'ramp_frontend': [
        os.path.join('templates', '*'),
        os.path.join('static', 'css', 'style.css'),
        os.path.join('static', 'css', 'themes', 'flat-blue.css'),
        os.path.join('static', 'img', '*'),
        os.path.join('static', 'img', 'backdrop', '*'),
        os.path.join('static', 'img', 'partners', '*'),
        os.path.join('static', 'img', 'powered_by', '*'),
        os.path.join('static', 'js', '*'),
        os.path.join('static', 'lib', 'css', '*'),
        os.path.join('static', 'lib', 'fonts', '*'),
        os.path.join('static', 'lib', 'img', '*'),
        os.path.join('static', 'lib', 'js', '*'),
    ]
}

setup(
    name=DISTNAME,
    maintainer=MAINTAINER,
    maintainer_email=MAINTAINER_EMAIL,
    description=DESCRIPTION,
    license=LICENSE,
    url=URL,
    version=VERSION,
    download_url=DOWNLOAD_URL,
    long_description=LONG_DESCRIPTION,
    zip_safe=False,  # the package can run out of an .egg file
    classifiers=CLASSIFIERS,
    packages=find_packages(),
    package_data=PACKAGE_DATA,
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
    entry_points={
        'console_scripts': ['ramp-frontend = ramp_frontend.cli:start']
    }
)
